@@ -119,8 +119,6 @@ import qualified Data.Text.Internal.Fusion as F
 
 -- | /Deprecated/.  Decode a 'ByteString' containing 7-bit ASCII
 -- encoded text.
---
--- This function is deprecated.  Use 'decodeLatin1' instead.
 decodeASCII :: ByteString -> Text
 decodeASCII = decodeUtf8
 {-# DEPRECATED decodeASCII "Use decodeUtf8 instead" #-}
@@ -269,6 +267,7 @@ streamDecodeUtf8With onErr = decodeChunk B.empty 0 0
                 UTF8_REJECT -> do
                   -- We encountered an encoding error
                   x <- peek curPtr'
+                  poke statePtr 0
                   case onErr desc (Just x) of
                     Nothing -> loop $ curPtr' `plusPtr` 1
                     Just c -> do
@@ -276,7 +275,6 @@ streamDecodeUtf8With onErr = decodeChunk B.empty 0 0
                       w <- unsafeSTToIO $
                            unsafeWrite dest (fromIntegral destOff) (safe c)
                       poke destOffPtr (destOff + fromIntegral w)
-                      poke statePtr 0
                       loop $ curPtr' `plusPtr` 1
 
                 _ -> do
@@ -288,7 +286,7 @@ streamDecodeUtf8With onErr = decodeChunk B.empty 0 0
                       return $! text arr 0 (fromIntegral n)
                   lastPtr <- peek curPtrPtr
                   let left = lastPtr `minusPtr` curPtr
-                      undecoded = case state of
+                      !undecoded = case state of
                         UTF8_ACCEPT -> B.empty
                         _           -> B.append undecoded0 (B.drop left bs)
                   return $ Some chunkText undecoded
